#include "LexStream.h"
#include "control.h"
#include "code.h"
#include "tab.h"

#include <sys/stat.h>
#include <iostream>
using namespace std;

static const char *TK_DEFINE_KEY_name           = "%Define",
                  *TK_MACRO_NAME_name           = "TK_MACRO_NAME",
                  *TK_SYMBOL_name               = "TK_SYMBOL",
                  *TK_BLOCK_name                = "TK_BLOCK",
                  *TK_TERMINALS_KEY_name        = "%Terminals",
                  *TK_KEYWORDS_KEY_name         = "%Keywords",
                  *TK_EOL_KEY_name              = "%Eol",
                  *TK_EOF_KEY_name              = "%Eof",
                  *TK_ERROR_KEY_name            = "%Error",
                  *TK_IDENTIFIER_KEY_name       = "%Identifier",
                  *TK_ALIAS_KEY_name            = "%Alias",
                  *TK_START_KEY_name            = "%Start",
                  *TK_RULES_KEY_name            = "%Rules",
                  *TK_EQUIVALENCE_name          = "::=",
                  *TK_PRIORITY_EQUIVALENCE_name = "::=?",
                  *TK_ARROW_name                = "->",
                  *TK_PRIORITY_ARROW_name       = "->?",
                  *TK_NAMES_KEY_name            = "%Names",
                  *TK_END_KEY_name              = "%End",
                  *TK_OR_MARKER_name            = "|",
                  *TK_HEADERS_KEY_name          = "%Headers",
                  *TK_TRAILERS_KEY_name         = "%Trailers",
                  *TK_EOF_name                  = "%Eof",
                  *NULL_STRING                  = "";

const char *LexStream::KeywordName(int kind)
{
    switch(kind)
    {
        case TK_DEFINE_KEY:                 return TK_DEFINE_KEY_name;
        case TK_MACRO_NAME:                 return TK_MACRO_NAME_name;
        case TK_SYMBOL:                     return TK_SYMBOL_name;
        case TK_BLOCK:                      return TK_BLOCK_name;
        case TK_TERMINALS_KEY:              return TK_TERMINALS_KEY_name;
        case TK_KEYWORDS_KEY:               return TK_KEYWORDS_KEY_name;
        case TK_EOL_KEY:                    return TK_EOL_KEY_name;
        case TK_EOF_KEY:                    return TK_EOF_KEY_name;
        case TK_ERROR_KEY:                  return TK_ERROR_KEY_name;
        case TK_IDENTIFIER_KEY:             return TK_IDENTIFIER_KEY_name;
        case TK_ALIAS_KEY:                  return TK_ALIAS_KEY_name;
        case TK_START_KEY:                  return TK_START_KEY_name;
        case TK_RULES_KEY:                  return TK_RULES_KEY_name;
        case TK_EQUIVALENCE:                return TK_EQUIVALENCE_name;
        case TK_PRIORITY_EQUIVALENCE:       return TK_PRIORITY_EQUIVALENCE_name;
        case TK_ARROW:                      return TK_ARROW_name;
        case TK_PRIORITY_ARROW:             return TK_PRIORITY_ARROW_name;
        case TK_NAMES_KEY:                  return TK_NAMES_KEY_name;
        case TK_END_KEY:                    return TK_END_KEY_name;
        case TK_OR_MARKER:                  return TK_OR_MARKER_name;
        case TK_HEADERS_KEY:                return TK_HEADERS_KEY_name;
        case TK_TRAILERS_KEY:               return TK_TRAILERS_KEY_name;
        case TK_EOF:                        return TK_EOF_name;

        default:                            break;
    }

    return NULL_STRING;
}


InputFileSymbol *LexStream::FindOrInsertFile(Tuple<const char *> &search_directory, const char *filename)
{
    //
    // Look in the search directory path for the filename.
    //
    char *full_filename = NULL;
    int filename_length = strlen(filename);
    for (int i = 0; i < search_directory.Length(); i++)
    {
        const char *directory_name = search_directory[i];
        int directory_name_length = strlen(directory_name),
            length = directory_name_length + filename_length;
        full_filename = new char[length + 2];
        strcpy(full_filename, directory_name);
        int index = directory_name_length;
        if (directory_name_length > 0 &&
            directory_name[directory_name_length - 1] != '\\' &&
            directory_name[directory_name_length - 1] != '/')
            full_filename[index++] = '/';
        if (*filename == '<') // remove brackets from bracketed symbols: <...>
        {
            strncpy(&(full_filename[index]), filename + 1, filename_length - 2);
            full_filename[index + (filename_length - 2)] = '\0';
        }
        else strcpy(&(full_filename[index]), filename);
        struct stat status;
        if (stat(full_filename, &status) == 0)
            break;
        delete [] full_filename;
        full_filename = NULL;
    }

    InputFileSymbol *file_symbol = NULL;
    if (full_filename != NULL)
    {
        //
        // No need to normalize full_filename as if two separate
        // names are used to access the same file, eventually, we
        // will have to revisit one of the names again. Therefore,
        // at worst, it may take a bit longer to detect the loop,
        // but sooner or later, it will be detected.
        //
        file_symbol = file_table.FindOrInsertName(full_filename, strlen(full_filename));
    }

    delete [] full_filename;

    return file_symbol;
} 

unsigned LexStream::FindLine(Tuple<unsigned> &line_location, unsigned location)
{
    int lo = 0,
        hi = line_location.Length() - 1;

    if (hi == 0)
        return 0;

    //
    // we can place the exit test at the bottom of the loop
    // since the line_location array will always contain at least
    // one element.
    //
    do
    {
        int mid = (lo + hi) / 2;

        if (line_location[mid] == location)
            return mid;
        if (line_location[mid] < location)
             lo = mid + 1;
        else hi = mid - 1;
    } while (lo < hi);

    return (line_location[lo] > location ? lo - 1 : lo);
}


//
//
//
void LexStream::ReportErrors()
{
    Tuple <const char *> msg;
    for (int i = 0; i < bad_tokens.Length(); i++)
    {
        msg.Reset();
        int token_index = bad_tokens[i].token_index;
        switch(bad_tokens[i].msg_code)
        {
            case NO_INPUT:
                 msg.Next() = "No input file specified";
                 break;
            case NO_TEMPLATE:
                 msg.Next() = "The template file ";
                 msg.Next() = option -> template_name;
                 msg.Next() = " could not be read.";
                 break;
            case NO_INCLUDE:
                 msg.Next() = "The include file ";
                 msg.Next() = NameString(token_index);
                 msg.Next() = " could not be read.";
                 break;
            case RECURSIVE_INCLUDE:
                 msg.Next() = "Attempt to recursively include file ";
                 msg.Next() = NameString(token_index);
                 break;
            case BAD_UNICODE:
                 msg.Next() = "Symbol \"";
                 msg.Next() = NameString(token_index);
                 msg.Next() = "\" contains an invalid unicode character.";
                 break;
            case BAD_OCTAL_ASCII_CODE:
                 msg.Next() = "Symbol \"";
                 msg.Next() = NameString(token_index);
                 msg.Next() = "\" contains an invalid ASCII octal code.";
                 break;
            case ISOLATED_BACKSLASH:
                 msg.Next() = "Symbol \"";
                 msg.Next() = NameString(token_index);
                 msg.Next() = "\" contains an isolated backslash.";
                 break;
            case UNDELIMITED_STRING_SYMBOL:
                 msg.Next() = "Quoted symbol not properly delimited. The closing quote must be followed by a space";
                 break;
            case UNTERMINATED_STRING_SYMBOL:
                 msg.Next() = "Quoted symbol not properly terminated";
                 break;
            case UNTERMINATED_BRACKET_SYMBOL:
                 msg.Next() = "Symbol enclosed in angular brackets not properly terminated";
                 break;
            case UNTERMINATED_BLOCK:
                 msg.Next() = "Block not properly terminated";
                 break;
            default:
                 assert(false);
        }

        EmitHeader("", token_index);
        for (int i = 0; i < msg.Length(); i++)
            option -> report.Put(msg[i]);
        option -> report.PutChar('\n');
        option -> report.Flush(stdout);
    }

    return;
}


void LexStream::EmitHeader(const char *header, LexStream::TokenIndex index)
{
    option -> report.Put(FileName(index));
    option -> report.Put(":");
    option -> report.Put(Line(index));
    option -> report.Put(":");
    option -> report.Put(Column(index));
    option -> report.Put(":");
    option -> report.Put(EndLine(index));
    option -> report.Put(":");
    option -> report.Put(EndColumn(index));
    option -> report.Put(":");
    option -> report.Put(StartLocation(index));
    option -> report.PutChar(':');
    option -> report.Put(EndLocation(index));
    option -> report.Put(": ");

    if (*header != '\0')
        option -> report.Put(header);

    return;
}

void LexStream::Emit(const char *header, const char *msg, LexStream::TokenIndex index)
{
    EmitHeader(header, index);
    option -> report.Put(msg);
    option -> report.PutChar('\n');

    option -> report.Print(Control::syslis);
    option -> report.Flush(stdout);

    return;
}


void LexStream::Emit(const char *header, Tuple<const char *> msg, LexStream::TokenIndex index)
{
    EmitHeader(header, index);
    for (int i = 0; i < msg.Length(); i++)
        option -> report.Put(msg[i]);
    option -> report.PutChar('\n');

    option -> report.Print(Control::syslis);
    option -> report.Flush(stdout);

    return;
}
