#ifndef GRAMMAR_INCLUDED
#define GRAMMAR_INCLUDED

#include "util.h"
#include "tuple.h"
#include "symbol.h"
#include "option.h"
#include "scanner.h"
#include "parser.h"
#include "set.h"

#undef INFINITY // disable macro definition of INFINITY

class Control;

class LCA;

class CTC;

class TTC;

class NTC;

class Grammar : public Util
{
public:

    enum { DEFAULT_SYMBOL = 0 };

    //
    // RULES is the structure that contain the rules of the grammar.
    // Every rule of the grammar is mapped into an integer, and given
    // rule, and we have access to a value RHS which is the index
    // location in the vector RHS where the right-hand-side of the rule
    // begins.  The right hand side of a certain rule represented by an
    // integer I starts at index location RULES[I].RHS in RHS, and
    // ends at index location RULES[I + 1].RHS - 1.  An extra
    // NUM_RULES + 1 element is used as a "fence" for the last rule.
    // The RHS vector as mentioned above is used to hold a complete
    // list of allthe right-hand-side symbols specified in the grammar.
    //
    class RuleElement
    {
    public:
        int lhs,
            separator_token,
            rhs_index,

            source_index; // the index in the parser object of this rule.
        //
        // In case an or delimiter, '|', is used, we may still want
        // to know what equivalence symbol started the chain.
        //
        int produces_token;

        bool IsPriorityProduction()
        {
            return produces_token == TK_PRIORITY_ARROW ||
                   produces_token == TK_PRIORITY_EQUIVALENCE;
        }
        bool IsArrowProduction()
        {
            return produces_token == TK_ARROW || produces_token == TK_PRIORITY_ARROW;
        }
        bool IsAlternateProduction() { return separator_token == TK_OR_MARKER; }
    };

    //
    //
    //
    class ClassnameElement
    {
    public:
        ClassnameElement() : specified_name(NULL),
                             real_name(NULL),
                             is_terminal_class(false),
                             needs_environment(false),
                             array_element_type_symbol(NULL)
        {}

        char *specified_name,
             *real_name;
        bool is_terminal_class,
             needs_environment;
        VariableSymbol *array_element_type_symbol;
        Tuple<int> rule,
                   ungenerated_rule,
                   interface,
                   rhs_type_index;
        SymbolLookupTable symbol_set;
    };

    //
    // 
    //
    struct ProcessedRuleElement
    {
        int image,
            token_index,  // points to the lex_stream
            type_index,   // points to a symbol image
            symbol_index, // points to a symbol element in a symbol_set in a ClassnameElement
            position;     // position of this symbol in the right-hand side of the rule
    };

private:

    //
    // 
    //
    struct SymbolElement
    {
        VariableSymbol *symbol;
        int external_name_index;

        char *Name() { return symbol -> Name(); }
    };

    //
    //
    //
    struct ActionBlockElement
    {
        enum
        {
            INITIALIZE,
            BODY,
            FINALIZE
        };

        int rule_number,
            location;
        LexStream::TokenIndex block_token;
        TextBuffer *buffer;
    };

    //
    //
    //
    struct AttributeElement
    {
        int lhs_image;
        LexStream::TokenIndex block_token;
    };

    //
    //
    //
    class RuleAllocationElement
    {
    public:
        enum ListKind {
                        NOT_A_LIST,
                        EMPTY,
                        SINGLETON,
                        LEFT_RECURSIVE_EMPTY,
                        RIGHT_RECURSIVE_EMPTY,
                        LEFT_RECURSIVE_SINGLETON,
                        RIGHT_RECURSIVE_SINGLETON,
                        ADD_ELEMENT,
                        COPY_LIST
                      };

        RuleAllocationElement() : name(NULL),
                                  is_terminal_class(false),
                                  needs_environment(false),
                                  list_kind(NOT_A_LIST),
                                  list_symbol(0),
                                  list_position(0),
                                  element_symbol(0),
                                  element_position(0)
        {}

        char *name;
        bool is_terminal_class,
             needs_environment;
        ListKind list_kind;
        int list_symbol,
            list_position,
            element_type_symbol_index,
            element_symbol,
            element_position;
    };

    bool IsNullClassname(ClassnameElement &element)
    {
        return element.rule.Length() == 0 ||
               (strlen(element.real_name) == 0 &&
                (element.array_element_type_symbol == NULL ||
                 element.array_element_type_symbol -> NameLength() == 0));
    }

    TextBuffer notice_buffer;
    BitSet keyword_set,
           recover_set;
    SimpleMacroLookupTable local_macro_table,
                           rule_macro_table,
                           filter_macro_table,
                           export_macro_table;

    SimpleMacroSymbol *rule_number_macro,
                      *rule_text_macro,
                      *rule_size_macro,
                      *input_file_macro,
                      *current_line_macro,
                      *next_line_macro,
                      *symbol_declarations;

    SimpleMacroSymbol *InsertLocalMacro(const char *str, const char *value = NULL)
    {
        int length = strlen(str) + 1;
        char *macro_name = new char[length + 1];
        macro_name[0] = option -> escape;
        strcpy(&(macro_name[1]), str);

        SimpleMacroSymbol *macro_symbol = local_macro_table.InsertName(macro_name, length);
        if (value)
            macro_symbol -> SetValue(value);

        delete [] macro_name;

        return macro_symbol;
    }

    SimpleMacroSymbol *InsertLocalMacro(const char *macro_name, int value)
    {
        IntToString num(value);
        return InsertLocalMacro(macro_name, num.String());
    }

    SimpleMacroSymbol *FindLocalMacro(const char *str, int length)
    {
        char *macro_name = new char[length + 1];
        for (int i = 0; i < length; i++)
            macro_name[i] = Code::ToLower(str[i]);
        macro_name[length] = '\0';

        SimpleMacroSymbol *macro_symbol = local_macro_table.FindName(macro_name, length);

        delete [] macro_name;

        return macro_symbol;
    }

    SimpleMacroSymbol *InsertRuleMacro(const char *macro_name, char *value)
    {
        int length = strlen(macro_name);
        SimpleMacroSymbol *macro_symbol = rule_macro_table.InsertName(macro_name, length);
        assert(value);
        macro_symbol -> SetValue(value);

        return macro_symbol;
    }

    SimpleMacroSymbol *InsertRuleMacro(const char *macro_name, int value)
    {
        IntToString num(value);
        return InsertRuleMacro(macro_name, num.String());
    }

    SimpleMacroSymbol *InsertExportMacro(const char *str, int length)
    {
        char *macro_name = new char[length + 3];
        macro_name[0] = option -> escape;
        macro_name[1] = '_';
        strncpy(&(macro_name[2]), str, length);
        macro_name[length + 2] = '\0';

        SimpleMacroSymbol *macro_symbol = export_macro_table.FindOrInsertName(macro_name, length + 2);

        delete [] macro_name;

        return macro_symbol;
    }

    SimpleMacroSymbol *InsertFilterMacro(const char *name, char *value)
    {
        int length = strlen(name);
        char *macro_name = new char[length + 2];
        macro_name[0] = option -> escape;
        strncpy(&(macro_name[1]), name, length);
        macro_name[length + 1] = '\0';

        SimpleMacroSymbol *macro_symbol = filter_macro_table.FindOrInsertName(macro_name, length + 1);
        macro_symbol -> SetValue(value);

        delete [] macro_name;

        return macro_symbol;
    }

    //
    // The variables below are used to hold information about special
    // grammar symbols.
    //
    Tuple<VariableSymbol *> variable_symbol_pool;
    VariableSymbol *empty_symbol,
                   *identifier_symbol,
                   *eol_symbol,
                   *eof_symbol,
                   *error_symbol,
                   *accept_symbol,
                   *start_symbol,
                   *null_symbol;

    VariableSymbol *allocate_variable_symbol(char *keyword)
    {
        int length = strlen(keyword) + 1;
        char *string = new char[length + 1];

        string[0] = option -> escape;
        strcpy(string + 1, keyword);
        int i = variable_symbol_pool.NextIndex();
        variable_symbol_pool[i] = new VariableSymbol(string, length, 0, 0);

        delete [] string;

        return variable_symbol_pool[i];
    }

    VariableSymbol *GetSymbol(int);
    int GetSymbolIndex(int);
    int AssignSymbolIndex(VariableSymbol *);
    void ProcessExportedTerminals();
    void ProcessTerminals();
    void ProcessInitialAliases(Tuple<int> &);
    void ProcessRemainingAliases(Tuple<int> &);
    void ProcessTitleOrGlobalBlock(int, ActionBlockElement &);
    char *InsertInterface(SymbolLookupTable &, char *);
    void ProcessRules();
    void SetName(VariableSymbol *, VariableSymbol *, bool negate = false);
    void ProcessNames();

    Control *control;
    Option *option;
    Action *action_blocks;
    LexStream *lex_stream;
    VariableLookupTable *variable_table;
    MacroLookupTable *macro_table;

    const char *SkipMargin(TextBuffer *, const char *, const char *);

    void ProcessAstActions(Tuple<ActionBlockElement> &,
                           Tuple<ActionBlockElement> &,
                           Tuple<ActionBlockElement> &,
                           Array<const char *> &,
                           Tuple< Tuple<ProcessedRuleElement> > &,
                           SymbolLookupTable &,
                           Tuple<ClassnameElement> &);
    void CheckRecursivenessAndUpdate(Tuple<int> &, BoundedArray< Tuple<int> > &, Array<RuleAllocationElement> &);
    void CompleteClassnameInfo(LCA &,
                               TTC &,
                               BoundedArray< Tuple<int> > &,
                               Array<const char *> &,
                               Tuple< Tuple<ProcessedRuleElement> > &,
                               SymbolLookupTable &,
                               Tuple<ClassnameElement> &,
                               Array<RuleAllocationElement> &);
    void ProcessAstRule(ClassnameElement &, int, Tuple<ProcessedRuleElement> &);
    void ProcessAstMergedRules(LCA &, ClassnameElement &, Tuple<int> &, Tuple< Tuple<ProcessedRuleElement> > &);
    void ProcessCodeActions(Tuple<ActionBlockElement> &, Array<const char *> &, Tuple< Tuple<ProcessedRuleElement> > &);
    void ProcessActionBlock(ActionBlockElement &);
    void DisplayString(const char *, const char);
    void DisplaySymbol(const char *);
    void DisplayInput();

    //*********************************************************************************
    //*
    //* TODO: The functions below are placed here only temporarily so that I can debug
    //* them. They actually belong to the file Java.cpp. Similar functions need to be
    //* written for the other languages, C, C++, ...
    //*
    //*********************************************************************************
    void GenerateDefaultTitle(Tuple<ActionBlockElement> &);
    ActionFileSymbol *GenerateTitle(ActionFileLookupTable &, Tuple<ActionBlockElement> &, const char *, bool);
    ActionFileSymbol *GenerateTitleAndGlobals(ActionFileLookupTable &, Tuple<ActionBlockElement> &, const char *, bool);

    void GenerateVisitorHeaders(TextBuffer &, const char *, const char *);
    void GenerateVisitorMethods(NTC &, TextBuffer &, const char *, ClassnameElement &, BitSet &);
    void GenerateEqualsMethod(NTC &, TextBuffer &, const char *, ClassnameElement &, BitSet &);
    void GenerateHashcodeMethod(NTC &, TextBuffer &, const char *, ClassnameElement &, BitSet &);

    void GenerateSimpleVisitorInterface(TextBuffer &, const char *, const char *, SymbolLookupTable &);
    void GenerateArgumentVisitorInterface(TextBuffer &, const char *, const char *, SymbolLookupTable &);
    void GenerateResultVisitorInterface(TextBuffer &, const char *, const char *, SymbolLookupTable &);
    void GenerateResultArgumentVisitorInterface(TextBuffer &, const char *, const char *, SymbolLookupTable &);

    void GeneratePreorderVisitorInterface(TextBuffer &, const char *, const char *, SymbolLookupTable &);
    void GeneratePreorderVisitorAbstractClass(TextBuffer &, const char *, const char *, SymbolLookupTable &);

    void GenerateNoResultVisitorAbstractClass(TextBuffer &, const char *, const char *, SymbolLookupTable &);
    void GenerateResultVisitorAbstractClass(TextBuffer &, const char *, const char *, SymbolLookupTable &);


    void GenerateAstType(TextBuffer &, const char *, const char *);
    void GenerateAstListType(TextBuffer &, const char *, const char *);
    void GenerateAstTokenType(NTC &, TextBuffer &, const char *, const char *);
    void GenerateInterface(bool, TextBuffer &, const char *, const char *, Tuple<int> &, Tuple<int> &, Tuple<ClassnameElement> &);
    void GenerateCommentHeader(TextBuffer &, const char *, ClassnameElement &);
    void GenerateCommentHeader(TextBuffer &, Tuple<int> &);
    void ComputeInterfaces(ClassnameElement &, Array<const char *> &, Tuple<int> &);
    void GenerateListClass(CTC &, NTC &, TextBuffer &, const char *, ClassnameElement &, Array<const char *> &);
    void GenerateRuleClass(CTC &, NTC &, TextBuffer &, const char *, ClassnameElement &, Array<const char *> &);
    void GenerateMergedClass(CTC &,
                             NTC &,
                             TextBuffer &,
                             const char *,
                             ClassnameElement &,
                             Tuple< Tuple<ProcessedRuleElement> > &,
                             Array<const char *> &);
    void GenerateTerminalMergedClass(NTC &, TextBuffer &, const char *, ClassnameElement &, Array<const char *> &);
    void GenerateNullAstAllocation(Grammar *, TextBuffer &, int rule_no);
    void GenerateEnvironmentDeclaration(TextBuffer &, const char *);
    void GenerateListAllocation(CTC &ctc, Grammar *, TextBuffer &, int, RuleAllocationElement &);
    void GenerateAstAllocation(CTC &ctc, Grammar *, TextBuffer &, RuleAllocationElement &, Tuple<ProcessedRuleElement> &, Array<const char *> &, int);
    void GenerateTokenClass(TextBuffer &, const char *, const char *);

    char *ast_token_classname;
    int ast_token_interface;

public:
    Grammar(Control *control_,
            Action *action_blocks_,
            LexStream *lex_stream_,
            VariableLookupTable *variable_table_,
            MacroLookupTable *macro_table_);

    ~Grammar()
    {
        for (int i = 0; i < variable_symbol_pool.Length(); i++)
            delete variable_symbol_pool[i];
        return;
    }

    Parser parser;

    void Process();
    int num_items,
        num_symbols,
        num_single_productions,
        num_terminals,
        num_nonterminals,
        num_names,
        num_rules;

    int empty,
        identifier_image,
        eol_image,
        eof_image,
        error_image,
        accept_image,
        start_image;

    //
    // Buffer containing "notice" string to be output.
    //
    TextBuffer &NoticeBuffer() { return notice_buffer; }

    //
    // SYMBOL_INDEX is an array that maps symbol numbers to actual symbols.
    //
    Tuple<SymbolElement> symbol_index;
    Tuple<RuleElement> rules;
    Tuple<int> rhs_sym;
    Tuple<int> keywords;
    Tuple<int> recovers;
    Tuple<char *> name;
    Tuple<VariableSymbol *> exported_symbols;

    bool IsTerminal(int x)    { return x > 0 && x <= num_terminals; }
    bool IsNonTerminal(int x) { return x > num_terminals && x <= num_symbols; }
    bool IsKeyword(int x)     { return IsTerminal(x) && keyword_set[x]; }
    BitSet &KeywordSet()      { return keyword_set; }
    bool IsRecover(int x)     { return IsTerminal(x) && recover_set[x]; }
    BitSet &RecoverSet()      { return recover_set; }

    int RhsSize(int rule_no)  { return rules[rule_no + 1].rhs_index - rules[rule_no].rhs_index; }

    bool IsUnitProduction(int rule_no) { return rules[rule_no].IsArrowProduction() && RhsSize(rule_no) == 1; }

    char *RetrieveString(int i) { return symbol_index[i].Name(); }
    char *RetrieveName(int i)   { return name[i]; }

    int FirstRhsIndex(int rule_no) { return rules[rule_no].rhs_index; }
    int EndRhsIndex(int rule_no)   { return rules[(rule_no) + 1].rhs_index; }

    inline int FirstTerminal() { return 1; }
    inline int LastTerminal()  { return num_terminals; }

    inline int FirstNonTerminal() { return num_terminals + 1; }
    inline int LastNonTerminal()  { return num_symbols; }

    inline int FirstRule() { return 0; }
    inline int LastRule()  { return num_rules; }

    void RestoreSymbol(char *, char *);
    void PrintLargeToken(char *, char *, char *, int);
    void ReportLargeToken(char *, char *, char *, int);

    inline void GenerateCode(TextBuffer *, const char *, int);
    void ProcessActionLine(int, TextBuffer *, const char *, const char *, const char *, int, int);
    void ProcessMacro(TextBuffer *, char *, int);
};

class LCA
{
    BoundedArray< Tuple<int> > &ancestors;
    BoundedArray<BitSetWithOffset> ancestorClosure;
    BoundedArray<int> index_of;
    Stack<int> stack;

    enum { TABLE_SIZE = 509 }; // 509 is a prime
    Array< Tuple<int> > table;

    //
    // ComputeAncestorClosure computes ancestorClosure[nt] for some non-terminal NT using the
    // digraph algorithm.  ancestorClosure[NT] is the transitive closure of the ancestors
    // of NT.
    //
    void ComputeAncestorClosure(int nt)
    {
        //
        // ancestorClosure[NT] was initialized to { NT } for each non-terminal NT.
        //
        stack.Push(nt);
        int indx = stack.Length();
        index_of[nt] = indx;

        //
        // iterate over all items of NT
        //
        for (int i = 0; i < ancestors[nt].Length(); i++)
        {
            int ancestor = ancestors[nt][i];
            if (index_of[ancestor] == Util::OMEGA)
                ComputeAncestorClosure(ancestor);
            ancestorClosure[nt] += ancestorClosure[ancestor];
            index_of[nt] = Util::Min(index_of[nt], index_of[ancestor]);
        }

        if (index_of[nt] == indx)
        {
            for (int ancestor = stack.Top(); ancestor != nt; ancestor = stack.Top())
            {
                ancestorClosure[ancestor] = ancestorClosure[nt];
                index_of[ancestor] = Util::INFINITY;
                stack.Pop();
            }

            index_of[nt] = Util::INFINITY;
            stack.Pop();
            table[ancestorClosure[nt].Hash(TABLE_SIZE)].Next() = nt;
        }

        return;
    }

public:

    LCA(BoundedArray< Tuple<int> > &ancestors_) : ancestors(ancestors_),
                                                  table(TABLE_SIZE)
    {
        index_of.Resize(ancestors.Lbound(), ancestors.Ubound());
        index_of.Initialize(Util::OMEGA);
        ancestorClosure.Resize(ancestors.Lbound(), ancestors.Ubound());
        for (int nt = ancestors.Lbound(); nt <= ancestors.Ubound(); nt++)
        {
            ancestorClosure[nt].Initialize(ancestors.Size() + 1, ancestors.Lbound() - 1);
            ancestorClosure[nt].AddElement(nt);
        }
 
        for (int nt = ancestors.Lbound(); nt <= ancestors.Ubound(); nt++)
        {
            if (index_of[nt] == Util::OMEGA)
                ComputeAncestorClosure(nt);
        }
    }

    int Find(int nt1, int nt2)
    {
        if (nt1 != 0 && nt2 != 0)
        {
            BitSetWithOffset lca_set(ancestorClosure[nt1]);
            lca_set *= ancestorClosure[nt2];
            int hash_index = lca_set.Hash(TABLE_SIZE);
            for (int i = 0; i < table[hash_index].Length(); i++)
            {
                int ancestor = table[hash_index][i];
                if (ancestorClosure[ancestor] == lca_set)
                    return ancestor;
            }
        }

        return 0;
    }

};


//
// CTC (Class transitive closure)
//
// For each interface I associated with a nonterminal, compute the 
// transitive closure of the set of classes that may implement this
// interface.
//
class CTC
{
    BoundedArray< Tuple<int> > extends;
    BoundedArray< Tuple<int> > interface_map;

    BoundedArray<BitSet> closure;
    BoundedArray<int> index_of;
    Stack<int> stack;

    Tuple<Grammar::ClassnameElement> &classname;
    Array<const char *> &typestring;

    void ComputeClosure(int nt)
    {
        stack.Push(nt);
        int indx = stack.Length();
        index_of[nt] = indx;

        for (int i = 0; i < extends[nt].Length(); i++)
        {
            int dependent = extends[nt][i];
            if (index_of[dependent] == Util::OMEGA)
                ComputeClosure(dependent);
            closure[nt] += closure[dependent];
            index_of[nt] = Util::Min(index_of[nt], index_of[dependent]);
        }

        if (index_of[nt] == indx)
        {
            for (int dependent = stack.Top(); dependent != nt; dependent = stack.Top())
            {
                closure[dependent] = closure[nt];
                index_of[dependent] = Util::INFINITY;
                stack.Pop();
            }

            index_of[nt] = Util::INFINITY;
            stack.Pop();
        }

        return;
    }

public:

    CTC(Tuple<Grammar::ClassnameElement> &classname_, Array<const char *> &typestring_, BoundedArray< Tuple<int> > &is_extended_by)
        : classname(classname_),
          typestring(typestring_)
    {
        //
        // Reverse the original extension map.
        //
        extends.Resize(is_extended_by.Lbound(), is_extended_by.Ubound());
        for (int i = is_extended_by.Lbound(); i <= is_extended_by.Ubound(); i++)
        {
            for (int k = 0; k < is_extended_by[i].Length(); k++)
                extends[is_extended_by[i][k]].Next() = i;
        }

        index_of.Resize(extends.Lbound(), extends.Ubound());
        index_of.Initialize(Util::OMEGA);

        closure.Resize(extends.Lbound(), extends.Ubound());
        for (int nt = closure.Lbound(); nt <= closure.Ubound(); nt++)
            closure[nt].Initialize(classname.Length());

        for (int i = 0; i < classname.Length(); i++)
        {
            Tuple<int> &interface = classname[i].interface;
            for (int k = 0; k < interface.Length(); k++)
                closure[interface[k]].AddElement(i);
        }

        for (int nt = closure.Lbound(); nt <= closure.Ubound(); nt++)
        {
            if (index_of[nt] == Util::OMEGA)
                ComputeClosure(nt);
        }

        interface_map.Resize(extends.Lbound(), extends.Ubound());
        for (int i = interface_map.Lbound(); i <= interface_map.Ubound(); i++)
        {
            for (int k = 0; k < closure[i].Size(); k++)
            {
                if (closure[i][k])
                    interface_map[i].Next() = k;
            }
        }
    }

    BoundedArray< Tuple<int> > &GetInterfaceMap() { return interface_map; }

    bool IsTerminalClass(int interface)
    {
        bool is_terminal_class = true;
        for (int k = 0; k < interface_map[interface].Length(); k++)
        {
            int class_index = interface_map[interface][k];
            is_terminal_class = is_terminal_class && classname[class_index].is_terminal_class;
        }

        return is_terminal_class;
    }

    char *FindUniqueTypeFor(int interface)
    {
        if (interface < interface_map.Lbound()) // a terminal symbol?
             return classname[0].real_name; //ast_token_classname
        else if (interface_map[interface].Length() == 1)
             return classname[interface_map[interface][0]].real_name;
        else return NULL;
    }

    const char *FindBestTypeFor(int interface)
    {
        if (interface == 0) // The default (root) Ast
             return typestring[0];
        else if (interface < interface_map.Lbound()) // a terminal symbol?
             return classname[0].real_name; //ast_token_classname
        else if (interface_map[interface].Length() == 1)
             return classname[interface_map[interface][0]].real_name;
        else return typestring[interface];
    }
};


//
// TTC (Terminal transitive closure)
//
// A symbol is identified as terminal if either:
//     1. It is a terminal symbol
//     2. It is a nonterminal that only produces a single terminal 
//        in each of its rules.
//
class TTC
{
    Array<bool> is_terminal;
    Array<int> index_of;
    Stack<int> stack;

    BoundedArray< Tuple<int> > &global_map;
    Tuple< Tuple<Grammar::ProcessedRuleElement> > &processed_rule_map;

    //
    // ComputeClosure computes closure[nt] for some non-terminal NT using the
    // digraph algorithm.
    //
    void ComputeClosure(int nt)
    {
        stack.Push(nt);
        int indx = stack.Length();
        index_of[nt] = indx;

        is_terminal[nt] = true;

        for (int i = 0; i < global_map[nt].Length(); i++)
        {
            int rule_no = global_map[nt][i];
            if (processed_rule_map[rule_no].Length() != 1)
                is_terminal[nt] = false;
            else
            {
                int dependent = processed_rule_map[rule_no][0].image;

                if (index_of[dependent] == Util::OMEGA)
                    ComputeClosure(dependent);
                is_terminal[nt] = is_terminal[nt] && is_terminal[dependent];
                index_of[nt] = Util::Min(index_of[nt], index_of[dependent]);
            }
        }

        if (index_of[nt] == indx)
        {
            for (int dependent = stack.Top(); dependent != nt; dependent = stack.Top())
            {
                is_terminal[dependent] = is_terminal[nt];
                index_of[dependent] = Util::INFINITY;
                stack.Pop();
            }

            index_of[nt] = Util::INFINITY;
            stack.Pop();
        }

        return;
    }

public:

    TTC(BoundedArray< Tuple<int> > &global_map_, Tuple< Tuple<Grammar::ProcessedRuleElement> > &processed_rule_map_)
        : is_terminal(global_map_.Ubound() + 1),
          index_of(global_map_.Ubound() + 1, Util::OMEGA),
          global_map(global_map_),
          processed_rule_map(processed_rule_map_)
    {
        //
        // Symbol 0 is associated with the root type and is never
        // considered to be a terminal
        //
        is_terminal[0] = false;
        index_of[0] = Util::INFINITY;

        //
        // Identify each terminal symbol as a terminal.
        //
        for (int t = 1; t < global_map_.Lbound(); t++)
        {
            is_terminal[t] = true;
            index_of[t] = Util::INFINITY;
        }

        //
        // Calculate which nonterminal symbols only produce terminal singletons.
        //
        for (int nt = global_map.Lbound(); nt <= global_map.Ubound(); nt++)
        {
            if (index_of[nt] == Util::OMEGA)
                ComputeClosure(nt);
        }
    }

    bool IsTerminalSymbol(int i) { return is_terminal[i]; }
};


//
// NTC (Null transitive closure)
//
// A null ast can be produced for a nonterminal if either:
//     1. It produces $empty
//     2. It produces (in a single production) a nonterminal that is null
//
class NTC
{
    BoundedArray<bool> is_null;
    BoundedArray<int> index_of;
    Stack<int> stack;

    Grammar *grammar;
    BoundedArray< Tuple<int> > &global_map;

    //
    // ComputeClosure computes closure[nt] for some non-terminal NT using the
    // digraph algorithm.
    //
    void ComputeClosure(int nt)
    {
        stack.Push(nt);
        int indx = stack.Length();
        index_of[nt] = indx;

        is_null[nt] = false;

        for (int i = 0; i < global_map[nt].Length(); i++)
        {
            int rule_no = global_map[nt][i],
                source_index = grammar -> rules[rule_no].source_index,
                array_index = grammar -> parser.rules[source_index].array_element_type_index;
            if (array_index == 0) // not a list rule
            {
                if (grammar -> RhsSize(rule_no) == 0)
                    is_null[nt] = true;
                else if (grammar -> RhsSize(rule_no) == 1)
                {
                    int dependent = grammar -> rhs_sym[grammar -> FirstRhsIndex(rule_no)];
                    if (grammar -> IsNonTerminal(dependent))
                    {
                        if (index_of[dependent] == Util::OMEGA)
                            ComputeClosure(dependent);
                        is_null[nt] = is_null[nt] || is_null[dependent];
                        index_of[nt] = Util::Min(index_of[nt], index_of[dependent]);
                    }
                }
            }
        }

        if (index_of[nt] == indx)
        {
            for (int dependent = stack.Top(); dependent != nt; dependent = stack.Top())
            {
                is_null[dependent] = is_null[nt];
                index_of[dependent] = Util::INFINITY;
                stack.Pop();
            }

            index_of[nt] = Util::INFINITY;
            stack.Pop();
        }

        return;
    }

public:

    NTC(BoundedArray< Tuple<int> > &global_map_, Grammar *grammar_) : global_map(global_map_)
    {
        grammar = grammar_;
        is_null.Resize(global_map_.Lbound(), global_map_.Ubound());
        index_of.Resize(global_map_.Lbound(), global_map_.Ubound());
        index_of.Initialize(Util::OMEGA);

        //
        // Calculate which nonterminal symbols only produce terminal singletons.
        //
        for (int nt = global_map_.Lbound(); nt <= global_map_.Ubound(); nt++)
        {
            if (index_of[nt] == Util::OMEGA)
                ComputeClosure(nt);
        }
    }

    bool CanProduceNullAst(int i) { return (grammar -> IsNonTerminal(i) ? is_null[i] : false); }
};
#endif

